/*
 * Generated by asn1c-0.9.29 (http://lionet.info/asn1c)
 * From ASN.1 module "PKIX1Explicit88"
 * 	found in "rfc5280-a.1.asn1"
 * 	`asn1c -Werror -fcompound-names -fwide-types -D asn1/asn1c -no-gen-PER -no-gen-example`
 */

#include "asn1/asn1c/AttributeType.h"

/*
 * This type is implemented using OBJECT_IDENTIFIER,
 * so here we adjust the DEF accordingly.
 */
static const ber_tlv_tag_t asn_DEF_AttributeType_tags_1[] = {
	(ASN_TAG_CLASS_UNIVERSAL | (6 << 2))
};
asn_TYPE_descriptor_t asn_DEF_AttributeType = {
	"AttributeType",
	"AttributeType",
	&asn_OP_OBJECT_IDENTIFIER,
	asn_DEF_AttributeType_tags_1,
	sizeof(asn_DEF_AttributeType_tags_1)
		/sizeof(asn_DEF_AttributeType_tags_1[0]), /* 1 */
	asn_DEF_AttributeType_tags_1,	/* Same as above */
	sizeof(asn_DEF_AttributeType_tags_1)
		/sizeof(asn_DEF_AttributeType_tags_1[0]), /* 1 */
	{ NULL, NULL, OBJECT_IDENTIFIER_constraint },
	NULL, 0,	/* No members */
	NULL	/* No specifics */
};
