#ifndef SRC_AS_NUMBER_H_
#define SRC_AS_NUMBER_H_

#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdint.h>

struct asn_range {
	uint32_t min;
	uint32_t max;
};

#endif /* SRC_AS_NUMBER_H_ */
