/*
 * Generated by asn1c-0.9.29 (http://lionet.info/asn1c)
 * From ASN.1 module "SignedObjectSyntax"
 * 	found in "rfc6488.asn1"
 * 	`asn1c -Werror -fcompound-names -fwide-types -D asn1/asn1c -no-gen-PER -no-gen-example`
 */

#ifndef	_BinaryTime_H_
#define	_BinaryTime_H_

#include "asn1/asn1c/INTEGER.h"

/* BinaryTime */
typedef INTEGER_t	 BinaryTime_t;

/* Implementation */
extern asn_TYPE_descriptor_t asn_DEF_BinaryTime;
asn_struct_free_f BinaryTime_free;
asn_struct_print_f BinaryTime_print;
asn_constr_check_f BinaryTime_constraint;
ber_type_decoder_f BinaryTime_decode_ber;
der_type_encoder_f BinaryTime_encode_der;
xer_type_encoder_f BinaryTime_encode_xer;

#endif	/* _BinaryTime_H_ */
