/*
 * Generated by asn1c-0.9.29 (http://lionet.info/asn1c)
 * From ASN.1 module "CryptographicMessageSyntax2004"
 * 	found in "rfc5652-12.1.asn1"
 * 	`asn1c -Werror -fcompound-names -fwide-types -D asn1/asn1c -no-gen-PER -no-gen-example`
 */

#ifndef	_ContentType_H_
#define	_ContentType_H_

#include "asn1/asn1c/OBJECT_IDENTIFIER.h"

/* ContentType */
typedef OBJECT_IDENTIFIER_t	 ContentType_t;

/* Implementation */
extern asn_TYPE_descriptor_t asn_DEF_ContentType;
asn_struct_free_f ContentType_free;
asn_struct_print_f ContentType_print;
asn_constr_check_f ContentType_constraint;
ber_type_decoder_f ContentType_decode_ber;
der_type_encoder_f ContentType_encode_der;
xer_type_encoder_f ContentType_encode_xer;

#endif	/* _ContentType_H_ */
