/*
 * Generated by asn1c-0.9.29 (http://lionet.info/asn1c)
 * From ASN.1 module "RPKIManifest"
 * 	found in "rfc6486-a.asn1"
 * 	`asn1c -Werror -fcompound-names -fwide-types -D asn1/asn1c -no-gen-PER -no-gen-example`
 */

#ifndef	_Manifest_H_
#define	_Manifest_H_

#include "asn1/asn1c/FileAndHash.h"
#include "asn1/asn1c/GeneralizedTime.h"
#include "asn1/asn1c/INTEGER.h"
#include "asn1/asn1c/OBJECT_IDENTIFIER.h"
#include "asn1/asn1c/asn_SEQUENCE_OF.h"

/* Manifest */
typedef struct Manifest {
	INTEGER_t	*version	/* DEFAULT 0 */;
	INTEGER_t	 manifestNumber;
	GeneralizedTime_t	 thisUpdate;
	GeneralizedTime_t	 nextUpdate;
	OBJECT_IDENTIFIER_t	 fileHashAlg;
	struct Manifest__fileList {
		A_SEQUENCE_OF(struct FileAndHash) list;
		
		/* Context for parsing across buffer boundaries */
		asn_struct_ctx_t _asn_ctx;
	} fileList;
	
	/* Context for parsing across buffer boundaries */
	asn_struct_ctx_t _asn_ctx;
} Manifest_t;

/* Implementation */
extern asn_TYPE_descriptor_t asn_DEF_Manifest;

#endif	/* _Manifest_H_ */
