/*
 * Generated by asn1c-0.9.29 (http://lionet.info/asn1c)
 * From ASN.1 module "PKIX1Explicit88"
 * 	found in "rfc5280-a.1.asn1"
 * 	`asn1c -Werror -fcompound-names -fwide-types -D asn1/asn1c -no-gen-PER -no-gen-example`
 */

#ifndef	_TBSCertList_H_
#define	_TBSCertList_H_

#include "asn1/asn1c/AlgorithmIdentifier.h"
#include "asn1/asn1c/CertificateSerialNumber.h"
#include "asn1/asn1c/Extensions.h"
#include "asn1/asn1c/Name.h"
#include "asn1/asn1c/Time.h"
#include "asn1/asn1c/Version.h"

/* Forward definitions */
typedef struct TBSCertList__revokedCertificates__Member {
	CertificateSerialNumber_t	 userCertificate;
	Time_t	 revocationDate;
	struct Extensions	*crlEntryExtensions	/* OPTIONAL */;
	
	/* Context for parsing across buffer boundaries */
	asn_struct_ctx_t _asn_ctx;
} TBSCertList__revokedCertificates__Member;

/* TBSCertList */
typedef struct TBSCertList {
	Version_t	*version	/* OPTIONAL */;
	AlgorithmIdentifier_t	 signature;
	Name_t	 issuer;
	Time_t	 thisUpdate;
	struct Time	*nextUpdate	/* OPTIONAL */;
	struct TBSCertList__revokedCertificates {
		A_SEQUENCE_OF(TBSCertList__revokedCertificates__Member) list;
		
		/* Context for parsing across buffer boundaries */
		asn_struct_ctx_t _asn_ctx;
	} *revokedCertificates;
	struct Extensions	*crlExtensions	/* OPTIONAL */;
	
	/* Context for parsing across buffer boundaries */
	asn_struct_ctx_t _asn_ctx;
} TBSCertList_t;

/* Implementation */
extern asn_TYPE_descriptor_t asn_DEF_TBSCertList;
extern asn_SEQUENCE_specifics_t asn_SPC_TBSCertList_specs_1;
extern asn_TYPE_member_t asn_MBR_TBSCertList_1[7];

#endif	/* _TBSCertList_H_ */
