/*
 * Generated by asn1c-0.9.29 (http://lionet.info/asn1c)
 * From ASN.1 module "CryptographicMessageSyntax2004"
 * 	found in "rfc5652-12.1.asn1"
 * 	`asn1c -Werror -fcompound-names -fwide-types -D asn1/asn1c -no-gen-PER -no-gen-example`
 */

#ifndef	_RevocationInfoChoice_H_
#define	_RevocationInfoChoice_H_

#include "asn1/asn1c/CertificateList.h"
#include "asn1/asn1c/OtherRevocationInfoFormat.h"

/* Dependencies */
typedef enum RevocationInfoChoice_PR {
	RevocationInfoChoice_PR_NOTHING,	/* No components present */
	RevocationInfoChoice_PR_crl,
	RevocationInfoChoice_PR_other
} RevocationInfoChoice_PR;

/* RevocationInfoChoice */
typedef struct RevocationInfoChoice {
	RevocationInfoChoice_PR present;
	union RevocationInfoChoice_u {
		CertificateList_t	 crl;
		OtherRevocationInfoFormat_t	 other;
	} choice;
	
	/* Context for parsing across buffer boundaries */
	asn_struct_ctx_t _asn_ctx;
} RevocationInfoChoice_t;

/* Implementation */
extern asn_TYPE_descriptor_t asn_DEF_RevocationInfoChoice;
extern asn_CHOICE_specifics_t asn_SPC_RevocationInfoChoice_specs_1;
extern asn_TYPE_member_t asn_MBR_RevocationInfoChoice_1[2];

#endif	/* _RevocationInfoChoice_H_ */
