/*
 * Generated by asn1c-0.9.29 (http://lionet.info/asn1c)
 * From ASN.1 module "CryptographicMessageSyntax2004"
 * 	found in "rfc5652-12.1.asn1"
 * 	`asn1c -Werror -fcompound-names -fwide-types -D asn1/asn1c -no-gen-PER -no-gen-example`
 */

#include "asn1/asn1c/SignerInfo.h"

asn_TYPE_member_t asn_MBR_SignerInfo_1[] = {
	{ ATF_NOFLAGS, 0, offsetof(struct SignerInfo, version),
		(ASN_TAG_CLASS_UNIVERSAL | (2 << 2)),
		0,
		&asn_DEF_CMSVersion,
		NULL,
		{ NULL, NULL, NULL },
		NULL, NULL, /* No default value */
		"version"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct SignerInfo, sid),
		-1 /* Ambiguous tag (CHOICE?) */,
		0,
		&asn_DEF_SignerIdentifier,
		NULL,
		{ NULL, NULL, NULL },
		NULL, NULL, /* No default value */
		"sid"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct SignerInfo, digestAlgorithm),
		(ASN_TAG_CLASS_UNIVERSAL | (16 << 2)),
		0,
		&asn_DEF_DigestAlgorithmIdentifier,
		NULL,
		{ NULL, NULL, NULL },
		NULL, NULL, /* No default value */
		"digestAlgorithm"
		},
	{ ATF_POINTER, 1, offsetof(struct SignerInfo, signedAttrs),
		(ASN_TAG_CLASS_CONTEXT | (0 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_SignedAttributes,
		NULL,
		{ NULL, NULL, NULL },
		NULL, NULL, /* No default value */
		"signedAttrs"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct SignerInfo, signatureAlgorithm),
		(ASN_TAG_CLASS_UNIVERSAL | (16 << 2)),
		0,
		&asn_DEF_SignatureAlgorithmIdentifier,
		NULL,
		{ NULL, NULL, NULL },
		NULL, NULL, /* No default value */
		"signatureAlgorithm"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct SignerInfo, signature),
		(ASN_TAG_CLASS_UNIVERSAL | (4 << 2)),
		0,
		&asn_DEF_SignatureValue,
		NULL,
		{ NULL, NULL, NULL },
		NULL, NULL, /* No default value */
		"signature"
		},
	{ ATF_POINTER, 1, offsetof(struct SignerInfo, unsignedAttrs),
		(ASN_TAG_CLASS_CONTEXT | (1 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_UnsignedAttributes,
		NULL,
		{ NULL, NULL, NULL },
		NULL, NULL, /* No default value */
		"unsignedAttrs"
		},
};
static const ber_tlv_tag_t asn_DEF_SignerInfo_tags_1[] = {
	(ASN_TAG_CLASS_UNIVERSAL | (16 << 2))
};
static const asn_TYPE_tag2member_t asn_MAP_SignerInfo_tag2el_1[] = {
    { (ASN_TAG_CLASS_UNIVERSAL | (2 << 2)), 0, 0, 0 }, /* version */
    { (ASN_TAG_CLASS_UNIVERSAL | (4 << 2)), 5, 0, 0 }, /* signature */
    { (ASN_TAG_CLASS_UNIVERSAL | (16 << 2)), 1, 0, 2 }, /* issuerAndSerialNumber */
    { (ASN_TAG_CLASS_UNIVERSAL | (16 << 2)), 2, -1, 1 }, /* digestAlgorithm */
    { (ASN_TAG_CLASS_UNIVERSAL | (16 << 2)), 4, -2, 0 }, /* signatureAlgorithm */
    { (ASN_TAG_CLASS_CONTEXT | (0 << 2)), 1, 0, 1 }, /* subjectKeyIdentifier */
    { (ASN_TAG_CLASS_CONTEXT | (0 << 2)), 3, -1, 0 }, /* signedAttrs */
    { (ASN_TAG_CLASS_CONTEXT | (1 << 2)), 6, 0, 0 } /* unsignedAttrs */
};
asn_SEQUENCE_specifics_t asn_SPC_SignerInfo_specs_1 = {
	sizeof(struct SignerInfo),
	offsetof(struct SignerInfo, _asn_ctx),
	asn_MAP_SignerInfo_tag2el_1,
	8,	/* Count of tags in the map */
	-1,	/* First extension addition */
};
asn_TYPE_descriptor_t asn_DEF_SignerInfo = {
	"SignerInfo",
	"SignerInfo",
	&asn_OP_SEQUENCE,
	asn_DEF_SignerInfo_tags_1,
	sizeof(asn_DEF_SignerInfo_tags_1)
		/sizeof(asn_DEF_SignerInfo_tags_1[0]), /* 1 */
	asn_DEF_SignerInfo_tags_1,	/* Same as above */
	sizeof(asn_DEF_SignerInfo_tags_1)
		/sizeof(asn_DEF_SignerInfo_tags_1[0]), /* 1 */
	{ NULL, NULL, SEQUENCE_constraint },
	asn_MBR_SignerInfo_1,
	7,	/* Elements count */
	&asn_SPC_SignerInfo_specs_1	/* Additional specs */
};
