/*
 * Generated by asn1c-0.9.29 (http://lionet.info/asn1c)
 * From ASN.1 module "CryptographicMessageSyntax2004"
 * 	found in "rfc5652-12.1.asn1"
 * 	`asn1c -Werror -fcompound-names -fwide-types -D asn1/asn1c -no-gen-PER -no-gen-example`
 */

#include "asn1/asn1c/SigningTime.h"

/*
 * This type is implemented using Time,
 * so here we adjust the DEF accordingly.
 */
asn_TYPE_descriptor_t asn_DEF_SigningTime = {
	"SigningTime",
	"SigningTime",
	&asn_OP_CHOICE,
	NULL,	/* No effective tags (pointer) */
	0,	/* No effective tags (count) */
	NULL,	/* No tags (pointer) */
	0,	/* No tags (count) */
	{ NULL, NULL, CHOICE_constraint },
	asn_MBR_Time_1,
	2,	/* Elements count */
	&asn_SPC_Time_specs_1	/* Additional specs */
};
