/*
 * Generated by asn1c-0.9.29 (http://lionet.info/asn1c)
 * From ASN.1 module "IPAddrAndASCertExtn"
 * 	found in "rfc3779.asn1"
 * 	`asn1c -Werror -fcompound-names -fwide-types -D asn1/asn1c -no-gen-PER -no-gen-example`
 */

#include "asn1/asn1c/IPAddressChoice.h"

#include "asn1/asn1c/constr_SEQUENCE_OF.h"

static asn_TYPE_member_t asn_MBR_addressesOrRanges_3[] = {
	{ ATF_POINTER, 0, 0,
		-1 /* Ambiguous tag (CHOICE?) */,
		0,
		&asn_DEF_IPAddressOrRange,
		NULL,
		{ NULL, NULL, NULL },
		NULL, NULL, /* No default value */
		""
		},
};
static const ber_tlv_tag_t asn_DEF_addressesOrRanges_tags_3[] = {
	(ASN_TAG_CLASS_UNIVERSAL | (16 << 2))
};
static asn_SET_OF_specifics_t asn_SPC_addressesOrRanges_specs_3 = {
	sizeof(struct IPAddressChoice__addressesOrRanges),
	offsetof(struct IPAddressChoice__addressesOrRanges, _asn_ctx),
	2,	/* XER encoding is XMLValueList */
};
static /* Use -fall-defs-global to expose */
asn_TYPE_descriptor_t asn_DEF_addressesOrRanges_3 = {
	"addressesOrRanges",
	"addressesOrRanges",
	&asn_OP_SEQUENCE_OF,
	asn_DEF_addressesOrRanges_tags_3,
	sizeof(asn_DEF_addressesOrRanges_tags_3)
		/sizeof(asn_DEF_addressesOrRanges_tags_3[0]), /* 1 */
	asn_DEF_addressesOrRanges_tags_3,	/* Same as above */
	sizeof(asn_DEF_addressesOrRanges_tags_3)
		/sizeof(asn_DEF_addressesOrRanges_tags_3[0]), /* 1 */
	{ NULL, NULL, SEQUENCE_OF_constraint },
	asn_MBR_addressesOrRanges_3,
	1,	/* Single element */
	&asn_SPC_addressesOrRanges_specs_3	/* Additional specs */
};

asn_TYPE_member_t asn_MBR_IPAddressChoice_1[] = {
	{ ATF_NOFLAGS, 0, offsetof(struct IPAddressChoice, choice.inherit),
		(ASN_TAG_CLASS_UNIVERSAL | (5 << 2)),
		0,
		&asn_DEF_NULL,
		NULL,
		{ NULL, NULL, NULL },
		NULL, NULL, /* No default value */
		"inherit"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct IPAddressChoice, choice.addressesOrRanges),
		(ASN_TAG_CLASS_UNIVERSAL | (16 << 2)),
		0,
		&asn_DEF_addressesOrRanges_3,
		NULL,
		{ NULL, NULL, NULL },
		NULL, NULL, /* No default value */
		"addressesOrRanges"
		},
};
static const asn_TYPE_tag2member_t asn_MAP_IPAddressChoice_tag2el_1[] = {
    { (ASN_TAG_CLASS_UNIVERSAL | (5 << 2)), 0, 0, 0 }, /* inherit */
    { (ASN_TAG_CLASS_UNIVERSAL | (16 << 2)), 1, 0, 0 } /* addressesOrRanges */
};
asn_CHOICE_specifics_t asn_SPC_IPAddressChoice_specs_1 = {
	sizeof(struct IPAddressChoice),
	offsetof(struct IPAddressChoice, _asn_ctx),
	offsetof(struct IPAddressChoice, present),
	sizeof(((struct IPAddressChoice *)0)->present),
	asn_MAP_IPAddressChoice_tag2el_1,
	2,	/* Count of tags in the map */
	-1	/* Extensions start */
};
asn_TYPE_descriptor_t asn_DEF_IPAddressChoice = {
	"IPAddressChoice",
	"IPAddressChoice",
	&asn_OP_CHOICE,
	NULL,	/* No effective tags (pointer) */
	0,	/* No effective tags (count) */
	NULL,	/* No tags (pointer) */
	0,	/* No tags (count) */
	{ NULL, NULL, CHOICE_constraint },
	asn_MBR_IPAddressChoice_1,
	2,	/* Elements count */
	&asn_SPC_IPAddressChoice_specs_1	/* Additional specs */
};
